balances_mock = [
    {"tokenAddress": None, "token": None, "balance": "171000000000000000"},
    {
        "tokenAddress": "0x70cBa46d2e933030E2f274AE58c951C800548AeF",
        "token": {
            "name": "Basic Attention Token",
            "symbol": "BAT",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x70cBa46d2e933030E2f274AE58c951C800548AeF.png",
        },
        "balance": "1000000000000000000",
    },
    {
        "tokenAddress": "0x4644A362B7041aff2CF792fD61671b9F75234C27",
        "token": {
            "name": " best tron wallet | tronscan | tronlink | bttc",
            "symbol": "tronAddress.org",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x4644A362B7041aff2CF792fD61671b9F75234C27.png",
        },
        "balance": "8888",
    },
    {
        "tokenAddress": "0x91056D4A53E1faa1A84306D4deAEc71085394bC8",
        "token": {
            "name": "CoW Protocol Token",
            "symbol": "COW",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x91056D4A53E1faa1A84306D4deAEc71085394bC8.png",
        },
        "balance": "7455447592176417847",
    },
    {
        "tokenAddress": "0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600",
        "token": {
            "name": "CoW Protocol Token",
            "symbol": "COW",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600.png",
        },
        "balance": "8045157924038386764",
    },
    {
        "tokenAddress": "0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60",
        "token": {
            "name": "Dai",
            "symbol": "DAI",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60.png",
        },
        "balance": "1000000000000000000000",
    },
    {
        "tokenAddress": "0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1",
        "token": {
            "name": "Elden Ring Token",
            "symbol": "ERT",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1.png",
        },
        "balance": "10000000000000000000000",
    },
    {
        "tokenAddress": "0xD14DD51BE76e6dC28756a442dF786845e03a1D60",
        "token": {
            "name": "ethAddress.io",
            "symbol": "ethAddress.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xD14DD51BE76e6dC28756a442dF786845e03a1D60.png",
        },
        "balance": "888888",
    },
    {
        "tokenAddress": "0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8",
        "token": {
            "name": "GiantETHLP",
            "symbol": "gETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8.png",
        },
        "balance": "1000000000000000",
    },
    {
        "tokenAddress": "0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532",
        "token": {
            "name": "Gnosis Token",
            "symbol": "GNO",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532.png",
        },
        "balance": "400000000000000000",
    },
    {
        "tokenAddress": "0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F",
        "token": {
            "name": "Liquid staked Ether 2.0",
            "symbol": "stETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F.png",
        },
        "balance": "30196852081566501",
    },
    {
        "tokenAddress": "0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1",
        "token": {
            "name": "RareTron.io",
            "symbol": "RareTron.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1.png",
        },
        "balance": "66666",
    },
    {
        "tokenAddress": "0x61fD3b6d656F39395e32f46E2050953376c3f5Ff",
        "token": {
            "name": "Safe Token",
            "symbol": "SAFE",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x61fD3b6d656F39395e32f46E2050953376c3f5Ff.png",
        },
        "balance": "10362070000000000000000",
    },
    {
        "tokenAddress": "0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947",
        "token": {
            "name": "Super ETH",
            "symbol": "ETHx",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947.png",
        },
        "balance": "5000000000000000",
    },
    {
        "tokenAddress": "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
        "token": {
            "name": "Uniswap",
            "symbol": "UNI",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984.png",
        },
        "balance": "21295518536254652",
    },
    {
        "tokenAddress": "0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C",
        "token": {
            "name": "USD Coin",
            "symbol": "USDC",
            "decimals": 6,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C.png",
        },
        "balance": "10000000",
    },
    {
        "tokenAddress": "0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb",
        "token": {
            "name": "USD Coin",
            "symbol": "USDC",
            "decimals": 6,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb.png",
        },
        "balance": "925089",
    },
    {
        "tokenAddress": "0x38d9639da06f7a726cAf40c2E1010ac88517A085",
        "token": {
            "name": "VanityTron.io",
            "symbol": "VanityTron.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x38d9639da06f7a726cAf40c2E1010ac88517A085.png",
        },
        "balance": "6666666",
    },
    {
        "tokenAddress": "0x1B809925ba90c541d895D19f0b7D70eE281a987F",
        "token": {
            "name": "VanityTRX.org",
            "symbol": "VanityTRX.org",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1B809925ba90c541d895D19f0b7D70eE281a987F.png",
        },
        "balance": "888888",
    },
    {
        "tokenAddress": "0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1",
        "token": {
            "name": "Wrapped Ether",
            "symbol": "WETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1.png",
        },
        "balance": "499812357969727",
    },
    {
        "tokenAddress": "0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6",
        "token": {
            "name": "Wrapped Ether",
            "symbol": "WETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6.png",
        },
        "balance": "5600000000000000",
    },
]
